%% tclp__clp_types.pl
%% Copyright (C) 2001,2002 INRIA Emmanuel Coquery

%% This library is free software; you can redistribute it and/or
%% modify it under the terms of the GNU Lesser General Public
%% License as published by the Free Software Foundation; either
%% version 2.1 of the License, or (at your option) any later version.

%% This library is distributed in the hope that it will be useful,
%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
%% Lesser General Public License for more details.

%% You should have received a copy of the GNU Lesser General Public
%% License along with this library; if not, write to the Free Software
%% Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%                                                                           %%
%%                                  TCLP                                     %%
%%                                                                           %%
%%         Emmanuel Coquery, projet Contraintes, INRIA Rocquencourt          %%
%%                                                                           %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% $Header: /home/beaune/contraintes/coquery/cvs/contraintes/tclpchr/src/tclp__clp_types.pl,v 1.14 2002/09/04 15:00:31 coquery Exp $

:- module(tclp__clp_types, [':<'/2,
                            fresh/1,
                            fresh_term/1,
                            init/0,
                            types_reinit/0,
                            getbounds/3,
                            check_constraint/1,
                            term_type_up/2,
                            term_type_down/2,
                            cyclic/1, cyclic_up/1, cyclic_down/1,
                            type_info/2,
                            get_greater_vars/2,
                            get_lesser_vars/2,
                            try_generalize_var/2
                            ]).

:- use_module(tclp__def).
:- use_module(tclp__errors).
:- use_module(tclp__type_order).
:- use_module(library(chr)).
:- use_module(library(ordsets)).

%% Verbose modes: debug_clp_types

handler type.

constraints
            % subtyping constraint
            (':<')/2  ,
            tclp__var_ineq/2 , tclp__hi/2, tclp__lo/2,
            % expressing type constraints over a type variable
            tclp__parameter/5,
            % updates the hiset and hibounds (resp. loset and lobound)
            % of a list of variables
            tclp__update_hiset/2, tclp__update_loset/2,
            % updates the lower (resp. greater) bound of some variables
            tclp__update_lobound/2, tclp__update_hibound/2,
            % stops constraints until updates are done
            tclp__sem/0,
            % store the set of original leaves of a variable
            tclp__original_up/2, tclp__original_down/2,
            % computes glbs and lubs of variables
            tclp__vGLB/3, tclp__vLUB/3,
            % make the break
            tclp__do_break/1
            .

operator(100,xfx,[':<']).

%%%%%%%%%%%%%%%%%%%%%%%%%
%% Solver's core rules %%
%%%%%%%%%%%%%%%%%%%%%%%%%

% when a break is needed, do it.

tclp__parameter(X, Hibound, Hiset, Loset, Lobound) \
tclp__do_break(X)
       <=> tclp__break(Lobound, Hibound).


% combinaison of two sets of constraints over the same variable
% generaly used after unification of two type variables
tclp__parameter(X, Hibound1, Hiset1, Loset1, Lobound1),
tclp__parameter(X, Hibound2, Hiset2, Loset2, Lobound2),
tclp__sem
       <=> verbose(debug_clp_types,'clp types: unifying ...'),
           list_to_ord_set(Hiset1,SHiset1),
           list_to_ord_set(Hiset2,SHiset2),
           list_to_ord_set(Loset1,SLoset1),
           list_to_ord_set(Loset2,SLoset2),
           ord_union(SHiset1, SHiset2, NH),
           ord_del_element(NH, X,NewHiset),
           ord_union(SLoset1, SLoset2, NL),
           ord_del_element(NL,X,NewLoset),
           tclp__update_hiset_s(NewLoset,NewHiset),
           tclp__update_loset_s(NewHiset,NewLoset),
           tclp__glb(Hibound1,Hibound2,NewHibound),
           tclp__lub(Lobound1,Lobound2,NewLobound),
           tclp__parameter(X,NewHibound,NewHiset,NewLoset,NewLobound),
           tclp__update_hibound_s(NewLoset,NewHibound),
           tclp__update_lobound_s(NewHiset,NewLobound),
           tclp__do_break(X),
           %tclp__break(NewLobound, NewHibound),
           verbose(debug_clp_types,'clp types: unifying ok'),
           %tclp__sem,
           %verbose(debug_clp_types,'clp types: unifying ok2').
           tclp__sem.


%X :< Y ==> chr:findall_constraints(_:<_,L),
%           length(L,N),
%           verbose(debug_clp_types,['clp types : adding ',
%                                    X :< Y,
%                                    ' with ',
%                                    N,
%                                    ' ineq'
%                                    ]),
%           true.

% posing constraints :<
% transform the generic constraint :< to more specific ones :
% between variables, or constraint a variable by a lower (resp. greater) bound
X :< Y , tclp__sem <=> var(X),var(Y) |
                 tclp__var_ineq(X,Y), tclp__sem.
X :< Y , tclp__sem <=> nonvar(X),var(Y) |
                 tclp__transinf(X,XFlat) , tclp__lo(XFlat,Y), tclp__sem.
X :< Y , tclp__sem <=> var(X), nonvar(Y) |
                 tclp__transup(Y,YFlat) , tclp__hi(X,YFlat), tclp__sem.
X :< Y , tclp__sem <=> nonvar(X), nonvar(Y) |
                 tclp__transinf(X,XFlat) , tclp__transup(Y,YFlat),
                 tclp__break(XFlat, YFlat), tclp__sem.

% X :< X 
tclp__var_ineq(X,X) <=> true.

% X :< Y , changes the constraints over X and Y
% (expressed by the tclp__parameter constraint)
tclp__parameter(X, HiboundX, HisetX, LosetX, LoboundX),
tclp__parameter(Y, HiboundY, HisetY, LosetY, LoboundY) \
tclp__var_ineq(X,Y)
       <=> true |
           verbose(debug_clp_types,['clp types: ',X,' < ',Y,' ']),
           list_to_ord_set(HisetX,SHisetX),
           (
             ord_member(Y,SHisetX) -> verbose(debug_clp_types,['clp types: ',X,' < ',Y,' already done']),true
                                % the constraint was already in the store
           ;
             (
               list_to_ord_set(LosetX, SLosetX),
               ord_add_element(SLosetX,X,Loset),
                                % the variables to add to Y and its hiset
               list_to_ord_set(HisetY,SHisetY),
               ord_add_element(SHisetY,Y,Hiset),
                                % the variables to add to X and its loset
               tclp__update_hiset_s(Loset, Hiset),
               tclp__update_loset_s(Hiset, Loset),
               tclp__update_hibound_s(Loset, HiboundY),
               tclp__update_lobound_s(Hiset, LoboundX),
                                % check compatibility of the bounds of X and Y
                                % one could check bound compatibility on one
                                % of X or Y since LoboundX < LoboundY
               tclp__break(LoboundX, HiboundY)
             ),
%             length(Hiset,NH), length(Loset,NL),
%             verbose(debug_clp_types,['hiset ',NH]),
%             verbose(debug_clp_types,['loset ',NL]),
             verbose(debug_clp_types,['clp types: ',X,' < ',Y,' ok'])
           ).
           

% X :< tau
tclp__parameter(X, HiboundX, HisetX, LosetX, LoboundX) \
tclp__hi(X,Hibound)
       <=> tclp__update_hibound_s([ X | LosetX ], Hibound),
           tclp__do_break(X).

% tau :< X
tclp__parameter(X, HiboundX, HisetX, LosetX, LoboundX) \
tclp__lo(Lobound,X)
       <=> tclp__update_lobound_s([ X | HisetX ], Lobound),
           tclp__do_break(X).

% change X's hibound and hiset
tclp__update_hiset(X, Hiset) ,
tclp__parameter(X, HiboundX, HisetX, LosetX, LoboundX)
       <=> % Hiset is already a ordered set
           list_to_ord_set(HisetX, SHisetX),
           ord_union(Hiset, SHisetX, NewHiset),
%       append(Hiset, HisetX, HS2), remove_duplicates(HS2, NewHiset),
           tclp__parameter(X, HiboundX, NewHiset, LosetX, LoboundX).

% change X's hibound 
tclp__update_hibound(X,Hibound) ,
tclp__parameter(X, HiboundX, HisetX, LosetX, LoboundX)
       <=> tclp__glb(HiboundX, Hibound, NewHibound),
           tclp__parameter(X, NewHibound, HisetX, LosetX, LoboundX).

% change X's loset
tclp__update_loset(X, Loset) ,
tclp__parameter(X, HiboundX, HisetX, LosetX, LoboundX)
       <=> % Loset is already a ordered set
           list_to_ord_set(LosetX, SLosetX),
           ord_union(Loset, SLosetX, NewLoset),
%           append(Loset, LosetX, LS2), remove_duplicates(LS2, NewLoset),
           tclp__parameter(X, HiboundX, HisetX, NewLoset, LoboundX).

% change X's lobound 
tclp__update_lobound(X,Lobound) ,
tclp__parameter(X, HiboundX, HisetX, LosetX, LoboundX)
       <=> tclp__lub(LoboundX, Lobound, NewLobound),
           tclp__parameter(X, HiboundX, HisetX, LosetX, NewLobound).


% compute a lub
%tclp__original_down(X,OrigX) \
%tclp__vLUB(X,X,LUB)
%       <=> list_to_ord_set(OrigX,OrigLUB),
%           tclp__LUBVar(OrigLUB,LUB).

% LUB(X,X) is X
tclp__vLUB(X,X,LUB) <=>  LUB = X.

% both X and Y are introduced variables
tclp__original_down(X,OrigX),
tclp__original_down(Y,OrigY) \
tclp__vLUB(X,Y,LUB)
       <=> list_to_ord_set(OrigX,SOrigX),
           list_to_ord_set(OrigY,SOrigY),
           ord_union(SOrigX,SOrigY,OrigLUB),
           tclp__LUBVar(OrigLUB,LUB).

% the second variable is an original
tclp__original_down(X,OrigX) \
tclp__vLUB(X,Y,LUB)
       <=> list_to_ord_set(OrigX,SOrigX),
           ord_add_element(SOrigX,Y,OrigLUB),
           tclp__LUBVar(OrigLUB,LUB).

% the first variable is an original : we switch the two variables
% to apply the preceeding rule
tclp__original_down(X,OrigX) \
tclp__vLUB(Y,X,LUB) <=> tclp__vLUB(X,Y,LUB).

% both variables are originals
tclp__vLUB(X,Y,LUB) <=>
       list_to_ord_set([X,Y],OrigLUB),
       tclp__LUBVar(OrigLUB,LUB).

% compute a glb
%tclp__original_up(X,OrigX) \
%tclp__vGLB(X,X,GLB)
%       <=> list_to_ord_set(OrigX,OrigGLB),
%           tclp__GLBVar(OrigGLB,GLB).

% GLB(X,X) is X
tclp__vGLB(X,X,GLB) <=> GLB = X.

% both X and Y are introduced variables
tclp__original_up(X,OrigX),
tclp__original_up(Y,OrigY) \
tclp__vGLB(X,Y,GLB)
       <=> list_to_ord_set(OrigX,SOrigX),
           list_to_ord_set(OrigY,SOrigY),
           ord_union(SOrigX,SOrigY,OrigGLB),
           tclp__GLBVar(OrigGLB,GLB).

% the second variable is an original
tclp__original_up(X,OrigX) \
tclp__vGLB(X,Y,GLB)
       <=> list_to_ord_set(OrigX,SOrigX),
           ord_add_element(SOrigX,Y,OrigGLB),
           tclp__GLBVar(OrigGLB,GLB).

% the first variable is an original: we switch the two variables
% to apply the preceeding rule
tclp__original_up(X,OrigX) \
tclp__vGLB(Y,X,GLB) <=> tclp__vGLB(X,Y,GLB).

% both variables are originals
tclp__vGLB(X,Y,GLB) <=>
       list_to_ord_set([X,Y],OrigGLB),
       tclp__GLBVar(OrigGLB,GLB).

%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Auxiliary predicates %%
%%%%%%%%%%%%%%%%%%%%%%%%%%

% calling constraints over a list of a variable

tclp__update_hiset_s([],_).
tclp__update_hiset_s([X|L],Hiset) :-
        tclp__update_hiset(X, Hiset),
        tclp__update_hiset_s(L, Hiset).

tclp__update_loset_s([],_).
tclp__update_loset_s([X|L], Loset) :-
        tclp__update_loset(X, Loset),
        tclp__update_loset_s(L, Loset).

tclp__update_hibound_s([],_).
tclp__update_hibound_s([X|L],Hibound) :-
        tclp__update_hibound(X,Hibound),
        tclp__update_hibound_s(L,Hibound).

tclp__update_lobound_s([],_).
tclp__update_lobound_s([X|L],Lobound) :-
        tclp__update_lobound(X,Lobound),
        tclp__update_lobound_s(L,Lobound).

% transforming a type term into a flat one with variables
% tclp__transup does it by constraining up variables

% a variable is always flat
tclp__transup(X,X) :- var(X),!.

% compound type
tclp__transup(Type,Flat) :-
        Type =.. [Constr|Args],
        tclp__transup_args(Args,FlatArgs),
        Flat =.. [Constr|FlatArgs].

% for args list
tclp__transup_args([],[]).
tclp__transup_args([T|L],[X|R]) :-
        (var(T) -> X=T ; (fresh(X) ,X :< T)),
        tclp__transup_args(L,R).

% tclp__transinf does it by constraining down variables

% a variable is always flat
tclp__transinf(X,X) :- var(X),!.

% compound type
tclp__transinf(Type,Flat) :-
        Type =.. [Constr|Args],
        tclp__transinf_args(Args,FlatArgs),
        Flat =.. [Constr|FlatArgs].

% for args list
tclp__transinf_args([],[]).
tclp__transinf_args([T|L],[X|R]) :-
        (var(T) -> X=T ; (fresh(X) ,T :< X)),
        tclp__transinf_args(L,R).

% creating new type variables
fresh(X) :- var(X), !,
        builtin_type(top,Top),
        builtin_type(bot,Bot),
        tclp__parameter(X,Top,[],[],Bot).

%% fresh_term(Term)
%% call fresh on all the variables of term
fresh_term(X) :- var(X), !, fresh(X).
fresh_term([]) :- !.
fresh_term([X|Rest]) :- !, fresh_term(X), fresh_term(Rest).
fresh_term(Term) :- Term =.. [ _ | Args ],!, fresh_term(Args).

% computing the variables corresponding to the lub (resp. glb) of an
% originals set

find_VAR(OrigToTest, TestVar,  Rest, Origs, Var) :-
        list_to_ord_set(OrigToTest, SOrigTT),
        ( ord_seteq(SOrigTT, Origs) -> TestVar = Var ;
            find_VAR(Rest, Origs, Var)).

find_VAR([(tclp__original_up(TestVar,OrigToTest) # _) | Rest],
         Origs, Var) :-
        find_VAR(OrigToTest, TestVar,  Rest, Origs, Var).

find_VAR([(tclp__original_down(TestVar,OrigToTest) # _) | Rest],
         Origs, Var) :-
        find_VAR(OrigToTest, TestVar,  Rest, Origs, Var).

%% this version unifies introduced variables that stand for the same set
%% of original variables

find_VAR2(OrigToTest, TestVar, Rest, Origs, Var, Found) :-
        list_to_ord_set(OrigToTest, SOrigTT),
        ( ord_seteq(SOrigTT, Origs) -> TestVar = Var, Found=yes ; true),
        find_VAR2(Rest, Origs, Var, Found).
find_VAR2([],_,_,Found) :- !,Found==yes.
find_VAR2([(tclp__original_up(TestVar,OrigToTest) # _) | Rest],
          Origs,Var, Found) :-
        find_VAR2(OrigToTest, TestVar, Rest, Origs,Var, Found).
find_VAR2([(tclp__original_down(TestVar,OrigToTest) # _) | Rest],
          Origs,Var, Found) :-
        find_VAR2(OrigToTest, TestVar, Rest, Origs,Var, Found).

find_VAR2(L,O,V) :- find_VAR2(L,O,V,_).

tclp__GLBVar(OrigGLB,GLB) :-
        % OrigGLB is sorted
        chr:findall_constraints(tclp__original_up(_,_), AllOrigs),
        (
          find_VAR(AllOrigs, OrigGLB, GLB) -> true ;
          (
            tclp__original_down(GLB,[]),
            tclp__original_up(GLB,OrigGLB),
            fresh(GLB),
            verbose(debug_clp_types,['clp types: created glb ',GLB,' for ',OrigGLB]),
            tclp__link_up(GLB,OrigGLB)
          )
        ).
        
tclp__link_up(_,[]).
tclp__link_up(X,[Y|Ys]) :-
        X :< Y,
        tclp__link_up(X,Ys).
        
tclp__LUBVar(OrigLUB,LUB) :-
        % OrigLUB is sorted
        chr:findall_constraints(tclp__original_down(_,_), AllOrigs),
        (
          find_VAR(AllOrigs, OrigLUB, LUB) -> true;
          (
            tclp__original_up(LUB,[]),
            tclp__original_down(LUB,OrigLUB),
            fresh(LUB),
            verbose(debug_clp_types,['clp types: created lub ',LUB,' for ',OrigLUB]),
            tclp__link_down(OrigLUB,LUB)
          )
        ).

tclp__link_down([],_).
tclp__link_down([X|Xs],Y) :-
        X :< Y,
        tclp__link_down(Xs,Y).

%% FIXME code for debugging, remove if rest is correct
%test_all_original([]).
%test_all_original([X|R]) :-
%        (
%          chr:find_constraint(tclp__original_down(A,L),_),A==X,
%          write('found illegal original lub set'(X,R,L)),nl
%        ->
%          throw('found illegal original lub set'([X|R]))
%        ;
%          test_all_original(R)
%        ).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Description of the type hierarchy %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% This is now done in tclp__type_order

% %types : top int pred list(.) bottom and pair

% % computation of greatest lower bounds
% tclp__glb(term,term,term) :- !.
% tclp__glb(term,int,int) :- !.
% tclp__glb(term,pred,pred) :- !.
% tclp__glb(term,list(T),list(T)) :- !.
% tclp__glb(term,pair(T1,T2),pair(T1,T2)) :- !.

% tclp__glb(int,term,int) :- !.
% tclp__glb(int,int,int) :- !.

% tclp__glb(pred,term,pred) :- !.
% tclp__glb(pred,pred,pred) :- !.

% tclp__glb(list(T),term,list(T)) :- !.
% tclp__glb(list(T),list(T2),list(T3)) :- !,tclp__vGLB(T,T2,T3).

% tclp__glb(pair(T1,T2),term,pair(T1,T2)) :- !.
% tclp__glb(pair(T1A,T1B),pair(T2A,T2B),pair(T3A,T3B)) :- !,
%         tclp__vGLB(T1A,T2A,T3A),tclp__vGLB(T1B,T2B,T3B).

% tclp__glb(A,B,_) :- !, 
% 	verbose(debug_clp_types,[glb(A,B), ' failed']), fail.

% % computation of lowest upper bounds
% tclp__lub(term,_,term) :- !.
% tclp__lub(_,term,term) :- !.

% tclp__lub(int,int,int) :- !.
% tclp__lub(int,pred,term) :- !.
% tclp__lub(int,list(_),term) :- !.
% tclp__lub(int,bottom,int) :- !.
% tclp__lub(int,pair(_,_),term) :- !.

% tclp__lub(pred,int,term) :- !.
% tclp__lub(pred,pred,pred) :- !.
% tclp__lub(pred,list(_),term) :- !.
% tclp__lub(pred,bottom,pred) :- !.
% tclp__lub(pred,pair(_,_),term) :- !.

% tclp__lub(list(_),int,term) :- !.
% tclp__lub(list(_),pred,term) :- !.
% tclp__lub(list(T),list(T2),list(T3)) :- !,tclp__vLUB(T,T2,T3).
% tclp__lub(list(T),bottom,list(T)) :- !.
% tclp__lub(list(_),pair(_,_),term) :- !.

% tclp__lub(pair(_,_),int,term) :- !.
% tclp__lub(pair(_,_),pred,term) :- !.
% tclp__lub(pair(_,_),list(_),term) :- !.
% tclp__lub(pair(A,B),bottom,pair(A,B)) :- !.
% tclp__lub(pair(T1A,T1B),pair(T2A,T2B),pair(T3A,T3B)) :- !,
% 	tclp__vLUB(T1A,T2A,T3A), tclp__vLUB(T1B,T2B,T3B).

% tclp__lub(bottom,T,T) :- !.
% tclp__lub(T,bottom,T) :- !.

% tclp__lub(A,B,_) :- !, 
% 	verbose(debug_clp_types,[lub(A,B),' failed']), fail.

% % type comparison
% tclp__break(term,term) :- !.

% tclp__break(int,term) :- !.
% tclp__break(int,int) :- !.

% tclp__break(pred,term) :- !.
% tclp__break(pred,pred) :- !.

% tclp__break(list(_),term) :- !.
% tclp__break(list(T1),list(T2)) :- !, T1 :< T2.

% tclp__break(pair(_,_),term) :- !.
% tclp__break(pair(A,B),pair(C,D)) :- !, A :< C, B :< D.

% tclp__break(bottom,_) :- !.

% tclp__break(A,B) :- !, 
%         verbose(debug_clp_types,[break(A,B), ' failed']), fail.

% %:- include('type_defs.pl').

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% (Re)Initialisation of the constraint store
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

init :- chr:findall_constraints(tclp__sem, []),!,
        tclp__sem.
init.

%% reinit
%% removes all the constraints in the type checker solver
remove_cstr(_ # Id) :- chr:remove_constraint(Id).
remove_cstrs([]).
remove_cstrs([X|L]) :-
        remove_cstr(X), remove_cstrs(L).
rem_constraints(Pattern) :-
        chr:findall_constraints(Pattern, Constraints),
        remove_cstrs(Constraints).
types_reinit :-            
        rem_constraints(':<'(_,_)),            
        rem_constraints(tclp__var_ineq(_,_)),
        rem_constraints(tclp__hi(_,_)),
        rem_constraints(tclp__lo(_,_)),
        rem_constraints(tclp__parameter(_,_,_,_,_)),
        rem_constraints(tclp__update_hiset(_,_)),
        rem_constraints(tclp__update_loset(_,_)), 
        rem_constraints(tclp__update_lobound(_,_)),
        rem_constraints(tclp__update_hibound(_,_)),
        rem_constraints(tclp__sem),
        rem_constraints(tclp__original_up(_,_)),
        rem_constraints(tclp__original_down(_,_)),
        rem_constraints(tclp__vGLB(_,_,_)),
        rem_constraints(tclp__vLUB(_,_,_)),
        rem_constraints(tclp__do_break(_)).

%% prints the constraints
print_constraints :-
        chr:findall_constraints(_, Constraints),
        print_list(Constraints).

print_list([]) :- nl.
print_list([C#_|L]) :-
        write(C),nl, print_list(L).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Test of satisfiability of some constraints without effective commit
%% (using backtracking)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% check(constraint)
%% Testing of a new constraint without commiting it to the current
%% constraint store
check_constraint(C) :-
        findall(R,(C,!,R=a),L),!,
        L=[a].

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Information on type variables
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%
%% Bounds infos %%
%%%%%%%%%%%%%%%%%%

%% Get the upper/lower bounds of a type variable
%% :- pred getbounds(type_variable, type_term, type_term).
getbounds(X,UB,LB) :-
        var(X),!,
        once((
              chr:findall_constraints(X, tclp__parameter(X,_,_,_,_),
                                      Constraints),
              Constraints = [tclp__parameter(X,UB,_,_,LB)#_|_]
             ;
              print_error(['Can''t get type constraints for ',X]),fail
             )).
getbounds(X,_,_) :-
        print_error(['Type variable was instanciated : ',X]),fail.

%% Get some infos about a type variable
%% :- pred type_info(type_variable, pair(type_term, type_term)).
type_info(X,LB-UB) :-
        cyclic(X),!,LB = UB, UB = cycle.
type_info(X,LB-UB) :-
        getbounds(X,UB1,LB1),!,
        UB1 =.. [ UTCons | UArgs ],
        u_type_infos(UArgs, UArgsInfos),
        UB =.. [ UTCons | UArgsInfos ],
        LB1 =.. [ LTCons | LArgs ],
        b_type_infos(LArgs, LArgsInfos),
        LB =.. [ LTCons | LArgsInfos ]
        ;
        print_error('type info failed'),fail.
type_info(_,_) :-
        print_error('get_bounds failed'),fail.

u_type_infos([],[]).
u_type_infos([X|Xs],[I|Is]) :-
        type_info(X,_-I),
        u_type_infos(Xs,Is).

b_type_infos([],[]).
b_type_infos([X|Xs],[I|Is]) :-
        type_info(X,I-_),
        b_type_infos(Xs,Is).

all_term_up([],[]).
all_term_up([X|Rest],[UB|BRest]) :-
        term_type_up(X,UB),
        all_term_up(Rest,BRest).

all_term_down([],[]).
all_term_down([X|Rest],[LB|BRest]) :-
        term_type_down(X,LB),
        all_term_down(Rest,BRest).

term_type_up(X,UB) :-
        (
          var(X)
        ->
          type_info(X,_-UB)
        ;
          X =.. [Head|Args],
          %write('a '),write([Head|Args]),nl,
          all_term_up(Args,ArgsB),
          %write('b '),write([Head|ArgsB]),nl,
          UB =.. [Head|ArgsB]%,
          %write('c '),write(UB),nl
        ).

term_type_down(X,LB) :-
        (
          var(X)
        ->
          type_info(X,LB-_)
        ;
          X =.. [Head|Args],
          all_term_down(Args,ArgsB),
          LB =.. [Head|ArgsB]
        ).

%%%%%%%%%%%%%%%%%%
%% Cycles infos %%
%%%%%%%%%%%%%%%%%%

%% Tests if there is a cycle in some term
cyclic_up(A) :-
        cyclic_up(A,[],[],Ups),
        ord_member(A,Ups).
cyclic_up(A,Done,Current,Result) :-
        (
          ord_member(A,Done)
        ->
          ord_add_element(Current,A,Result)
        ;
          getbounds(A,_,UB),
          UB =.. [ _ | UArgs ],
          ord_add_element(Done,A,Done2),
          cyclic_up2(UArgs,Done2,Current,Result)
        ).
cyclic_up2([],_,Result,Result).
cyclic_up2([A|Rest],Done,Current,Result) :-
        cyclic_up(A,Done,Current,Accu),
        cyclic_up2(Rest,Done,Accu,Result).

cyclic_down(A) :-
        cyclic_down(A,[],[],Downs),
        ord_member(A,Downs).
cyclic_down(A,Done,Current,Result) :-
        (
          ord_member(A,Done)
        ->
          ord_add_element(Current,A,Result)
        ;
          getbounds(A,LB,_),
          LB =.. [ _ | UArgs ],
          ord_add_element(Done,A,Done2),
          cyclic_down2(UArgs,Done2,Current,Result)
        ).
cyclic_down2([],_,Result,Result).
cyclic_down2([A|Rest],Done,Current,Result) :-
        cyclic_down(A,Done,Current,Accu),
        cyclic_down2(Rest,Done,Accu,Result).

cyclic(A) :- cyclic_up(A) ; cyclic_down(A).

%%%%%%%%%%%%%%%%%
%% Order infos %%
%%%%%%%%%%%%%%%%%

%% Vars are all the type variables Y such that X :< Y
get_greater_vars(X, Vars) :-
        chr:find_constraint(X,tclp__parameter(X, _, Vars, _, _),_),!.

%% Vars are all the type variables Y such that Y :< X
get_lesser_vars(X, Vars) :-
        chr:find_constraint(X,tclp__parameter(X, _, _, Vars, _),_),!.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Propagation of the generalisation of a variable
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

strict_member(_,[]) :- fail.
strict_member(X,[Y|_]) :- X == Y,!.
strict_member(X,[_|L]) :- strict_member(X,L).

%% Finds all variables that are linked to X
find_linked(X,Current,All) :-
        (
          strict_member(X,Current)
        ->
          Current = All
        ;
          get_greater_vars(X, Greater),!,
          find_all_linked(Greater, [ X | Current ], Inter),
          get_lesser_vars(X, Lesser),!,
          find_all_linked(Lesser, Inter, All),!
        ).
find_all_linked([],All,All).
find_all_linked([X|L],Current,All) :-
        find_linked(X,Current,Inter),
        find_all_linked(L,Inter,All).

%%% Try the generalisation of X to the constant Name
%try_generalize_var(X, Name, Current, Done) :-
%        (
%          ord_member(X,Current)
%        ->
%          Current = Done
%        ;
%%          verbose(debug_type_inference,[ 'trying generalize ',X,' to ',Name]),
%          X :< Name, Name :< X,!,
%          ord_add_element(Current,X,Inter1),
%          get_greater_vars(X, Greater),!,
%%          verbose(debug_type_inference,[ 'uppers are ',Greater ]),
%          try_generalize_all_vars(Greater, Name, Inter1, Inter2),
%          get_lesser_vars(X, Lesser),!,
%%          verbose(debug_type_inference,[ 'lesser are ',Lesser ]),
%          try_generalize_all_vars(Lesser, Name, Inter2, Done),!
%        ).

%try_generalize_all_vars([], _, Done,Done).
%try_generalize_all_vars([ X | Xs ], Name, Current, Done) :-
%        try_generalize_var(X,Name,Current,Inter),!,
%        try_generalize_all_vars(Xs,Name,Inter,Done).

try_generalize_var2(X,Name) :-
        X :< Name, Name :< X,!.

try_generalize_all_vars([],_).
try_generalize_all_vars([X|Rest],Name) :-
        once(try_generalize_var2(X,Name)),
        try_generalize_all_vars(Rest,Name).

try_generalize_var(X,Name) :-
        find_linked(X,[],Linked),
        length(Linked,Len),
        verbose(debug_type_inference,['found ',Len,' vars linked to ',X]),
        try_generalize_all_vars(Linked,Name).
